from .liburing_h cimport io_uring, io_uring_sqe, io_uring_cqe, io_uring_op, io_uring_probe
from .liburing_h cimport io_uring_queue_init, io_uring_submit, io_uring_get_sqe, io_uring_sqe_set_data, io_uring_peek_cqe, io_uring_cqe_seen, io_uring_queue_exit, io_uring_get_probe_ring, io_uring_free_probe, io_uring_opcode_supported, io_uring_register_eventfd
from .liburing_h cimport io_uring_prep_nop, io_uring_prep_read, io_uring_prep_write, io_uring_prep_statx, io_uring_prep_openat, io_uring_prep_close, io_uring_prep_renameat, io_uring_prep_unlinkat, io_uring_prep_accept, io_uring_prep_connect, io_uring_prep_recv, io_uring_prep_send, io_uring_prep_socket, io_uring_prep_splice, io_uring_prep_cancel

from .statx cimport statx
from .socket cimport sockaddr_in, sockaddr_in6, inet_pton, AF_INET, AF_INET6, AF_UNIX, AF_UNSPEC, htons

cdef unsigned int MAX_ACCEPT_ADDRLEN = 32

import os
import logging
import platform

cpdef str get_opcode_name(int opcode):
    names = ["IORING_OP_NOP", "IORING_OP_READV", "IORING_OP_WRITEV", "IORING_OP_FSYNC",
        "IORING_OP_READ_FIXED", "IORING_OP_WRITE_FIXED", "IORING_OP_POLL_ADD", "IORING_OP_POLL_REMOVE",
        "IORING_OP_SYNC_FILE_RANGE", "IORING_OP_SENDMSG", "IORING_OP_RECVMSG", "IORING_OP_TIMEOUT",
        "IORING_OP_TIMEOUT_REMOVE", "IORING_OP_ACCEPT", "IORING_OP_ASYNC_CANCEL", "IORING_OP_LINK_TIMEOUT",
        "IORING_OP_CONNECT", "IORING_OP_FALLOCATE", "IORING_OP_OPENAT", "IORING_OP_CLOSE",
        "IORING_OP_RSRC_UPDATE", "IORING_OP_STATX", "IORING_OP_READ", "IORING_OP_WRITE",
        "IORING_OP_FADVISE", "IORING_OP_MADVISE", "IORING_OP_SEND", "IORING_OP_RECV",
        "IORING_OP_OPENAT2", "IORING_OP_EPOLL_CTL", "IORING_OP_SPLICE", "IORING_OP_PROVIDE_BUFFERS",
        "IORING_OP_REMOVE_BUFFERS", "IORING_OP_TEE", "IORING_OP_SHUTDOWN", "IORING_OP_RENAMEAT",
        "IORING_OP_UNLINKAT", "IORING_OP_MKDIRAT", "IORING_OP_SYMLINKAT", "IORING_OP_LINKAT",
        "IORING_OP_MSG_RING", "IORING_OP_FSETXATTR", "IORING_OP_SETXATTR", "IORING_OP_FGETXATTR",
        "IORING_OP_GETXATTR", "IORING_OP_SOCKET", "IORING_OP_URING_CMD", "IORING_OP_SENDZC_NOTIF",
        "IORING_OP_LAST"]
    return names[opcode]

cdef class SockAddr:
    cdef unsigned short family
    cdef sockaddr_in ipv4
    cdef sockaddr_in6 ipv6
    cdef unsigned int size
    #cdef sockaddr_un unix TODO add unix socket support

    def __init__(self, type, address):
        if isinstance(address[0], str):
            address = (address[0].encode('utf-8'), address[1])
        self.family = type

        if type == AF_INET:
            self.ipv4.sin_family = AF_INET
            self.ipv4.sin_port = htons(address[1])
            inet_pton(AF_INET, address[0], &self.ipv4.sin_addr)
            self.size = sizeof(self.ipv4)
        elif type == AF_INET6:
            self.ipv6.sin6_family = AF_INET6
            self.ipv6.sin6_port = htons(address[1])
            inet_pton(AF_INET6, address[0], &self.ipv6.sin6_addr)
            self.size = sizeof(self.ipv6)
        else:
            raise ValueError("Address Family not supported")
        
    cdef void* data(self):
        if self.family == AF_INET:
            return &self.ipv4
        elif self.family == AF_INET6:
            return &self.ipv6






cdef class StatxResult:
    cdef statx data

    @property
    def st_mode(self):
        return self.data.stx_mode
    @property
    def st_ino(self):
        return self.data.stx_ino
    @property
    def st_dev(self):
        return os.makedev(self.data.stx_dev_major, self.data.stx_dev_minor)
    @property
    def st_nlink(self):
        return self.data.stx_nlink
    @property
    def st_uid(self):
        return self.data.stx_uid
    @property
    def st_gid(self):
        return self.data.stx_gid
    @property
    def st_size(self):
        return self.data.stx_size
    @property
    def st_atime(self):
        return self.data.stx_atime.tv_sec + self.data.stx_atime.tv_nsec / 1000000000.0
    @property
    def st_mtime(self):
        return self.data.stx_mtime.tv_sec + self.data.stx_mtime.tv_nsec / 1000000000.0
    @property
    def st_ctime(self):
        return self.data.stx_ctime.tv_sec + self.data.stx_ctime.tv_nsec / 1000000000.0
    @property
    def st_atime_ns(self):
        return self.data.stx_atime.tv_nsec
    @property
    def st_mtime_ns(self):
        return self.data.stx_mtime.tv_nsec
    @property
    def st_ctime_ns(self):
        return self.data.stx_ctime.tv_nsec
    @property
    def st_rdev(self):
        return os.makedev(self.data.stx_rdev_major, self.data.stx_rdev_minor)
    @property
    def st_blocks(self):
        return self.data.stx_blocks
    @property
    def st_flags(self):
        return self.data.stx_flags

class IoUringNotSupportedException(Exception):
    pass

cdef class IoUringCqe:
    cdef int res
    cdef unsigned int flags
    cdef object user_data

    def __init__(self, int res, unsigned int flags, object user_data):
        if not self.is_supported():
            raise IoUringNotSupportedException
        self.res = res
        self.flags = flags
        self.user_data = user_data

    @staticmethod
    def is_supported():
        if platform.system() == "Linux":
            # check if the kernel is recent enough
            version = platform.release().split(".")
            if int(version[0]) >= 5 and int(version[1]) >= 1:
                return True
        return False

    @staticmethod
    cdef object from_cqe(io_uring_cqe* cqe):
        cdef object user_data;
        if cqe.user_data == 0:
            user_data = None
        else:
            user_data = <object><void*>cqe.user_data
        return IoUringCqe(
            cqe.res, 
            cqe.flags, 
            user_data
        )
    
    @property
    def data(self):
        return self.user_data
    
    @property
    def result(self):
        return self.res

cdef class IoUring:
    cdef io_uring ring
    cdef io_uring_probe* probe
    cdef bint open

    def __init__(self, entries=256):
        io_uring_queue_init(entries, &self.ring, 0)
        self.probe = io_uring_get_probe_ring(&self.ring)
        self.open = 1

    cdef int __check(self, io_uring_op opcode):
        if self.open:
            logging.debug(f"Scheduled opcode: {get_opcode_name(opcode)}")
            return self.__check_opcode(opcode)
        return 0

    cpdef int register_eventfd(self, int fd):
        return io_uring_register_eventfd(&self.ring, fd)
        
    cdef int __check_opcode(self, int opcode):
        return io_uring_opcode_supported(self.probe, opcode)

    def get_probe(self):
        codes = {}
        for op in range(io_uring_op.IORING_OP_LAST + 1):
            codes[op] = self.__check_opcode(op)
        return codes

    cpdef close(self):
        if self.open:
            io_uring_queue_exit(&self.ring)
            io_uring_free_probe(self.probe)
            self.open = 0

    cdef io_uring_sqe* __get_sqe(self):
        return io_uring_get_sqe(&self.ring)

    cpdef object get_completion(self, timeout=0):
        cdef io_uring_cqe* cqe
        cdef IoUringCqe ret

        if io_uring_peek_cqe(&self.ring, &cqe) == 0:
            ret = IoUringCqe.from_cqe(cqe)
            io_uring_cqe_seen(&self.ring, cqe)
            return ret

    cdef set_data(self,io_uring_sqe* sqe, object data):
        io_uring_sqe_set_data(sqe, <void*>data)
    
    cpdef int schedule_nop(self, object user_data):
        if self.__check(io_uring_op.IORING_OP_NOP) == 0:
            return -1
        cdef io_uring_sqe* sqe = self.__get_sqe()
        if sqe == NULL:
            return -2
        io_uring_prep_nop(sqe)
        self.set_data(sqe, user_data)
        return 1

    cpdef int schedule_read(self,  object user_data, int fd, bytes buf, unsigned long offset=0):
        if self.__check(io_uring_op.IORING_OP_READ) == 0:
                return -1
        cdef char* c_buffer = buf
        cdef int length = len(buf)
        cdef io_uring_sqe* sqe = self.__get_sqe()
        if sqe == NULL:
            return -2

        io_uring_prep_read(sqe, fd, c_buffer, length, offset)
        self.set_data(sqe, user_data)
        return 1
        
    cpdef int schedule_write(self, object user_data, int fd, bytes buf, unsigned long offset=0):
        if self.__check(io_uring_op.IORING_OP_WRITE) == 0:
                return -1
        cdef char* c_buffer = buf
        cdef int length = len(buf)
        cdef io_uring_sqe* sqe = self.__get_sqe()
        if sqe == NULL:
            return -2

        io_uring_prep_write(sqe, fd, c_buffer, length, offset)
        self.set_data(sqe, user_data)
        return 1

    cpdef int schedule_statx(self, object user_data, int dfd, bytes path, StatxResult ref, int flags=0, unsigned int mask=0):
        if self.__check(io_uring_op.IORING_OP_STATX) == 0:
                return -1
        cdef char* c_path = path
        cdef statx stx_buf
        cdef io_uring_sqe* sqe = self.__get_sqe()
        if sqe == NULL:
            return -2

        io_uring_prep_statx(sqe, dfd, c_path, flags, mask, <statx*>&ref.data)
        self.set_data(sqe, user_data)
        return 1

    cpdef int schedule_openat(self, object user_data, int dfd, bytes path, int flags, int mode):
        if self.__check(io_uring_op.IORING_OP_OPENAT) == 0:
                return -1
        cdef char* c_path = path
        cdef io_uring_sqe* sqe = self.__get_sqe()
        if sqe == NULL:
            return -2

        io_uring_prep_openat(sqe, dfd, c_path, flags, mode)
        self.set_data(sqe, user_data)
        return 1

    cpdef int schedule_close(self, object user_data, int fd):
        if self.__check(io_uring_op.IORING_OP_CLOSE) == 0:
                return -1
        cdef io_uring_sqe* sqe = self.__get_sqe()
        if sqe == NULL:
            return -2

        io_uring_prep_close(sqe, fd)
        self.set_data(sqe, user_data)
        return 1

    cpdef int schedule_renameat(self, object user_data, int olddfd, bytes oldpath, int newdfd, bytes newpath, int flags):
        if self.__check(io_uring_op.IORING_OP_RENAMEAT) == 0:
                return -1
        cdef char* c_oldpath = oldpath
        cdef char* c_newpath = newpath
        cdef io_uring_sqe* sqe = self.__get_sqe()
        if sqe == NULL:
            return -2

        io_uring_prep_renameat(sqe, olddfd, c_oldpath, newdfd, c_newpath, flags)
        self.set_data(sqe, user_data)
        return 1

    cpdef int schedule_unlinkat(self, object user_data, int dfd, bytes path, int flags):
        if self.__check(io_uring_op.IORING_OP_UNLINKAT) == 0:
            return -1
        cdef char* c_path = path
        cdef io_uring_sqe* sqe = self.__get_sqe()
        if sqe == NULL:
            return -2

        io_uring_prep_unlinkat(sqe, dfd, c_path, flags)
        self.set_data(sqe, user_data)
        return 1

    cpdef int schedule_accept(self, object user_data, int sockfd, int flags=0):
        if self.__check(io_uring_op.IORING_OP_ACCEPT) == 0:
            return -1
        cdef io_uring_sqe* sqe = self.__get_sqe()
        if sqe == NULL:
            return -2
            
        cdef unsigned int zero = 0
        io_uring_prep_accept(sqe, sockfd, NULL, &zero,  flags)
        self.set_data(sqe, user_data)
        return 1

    cpdef int schedule_connect(self, object user_data, int sockfd, SockAddr addr):
        if self.__check(io_uring_op.IORING_OP_CONNECT) == 0:
            return -1
        cdef io_uring_sqe* sqe = self.__get_sqe()
        if sqe == NULL:
            return -2
        cdef unsigned int zero = 0
    
        io_uring_prep_connect(sqe, sockfd, addr.data(), addr.size)
        self.set_data(sqe, user_data)
        return 1

    cpdef int schedule_recv(self, object user_data, int sockfd, bytes buf, int flags=0):
        if self.__check(io_uring_op.IORING_OP_RECV) == 0:
            return -1
        cdef char* c_buffer = buf
        cdef int length = len(buf)
        cdef io_uring_sqe* sqe = self.__get_sqe()
        if sqe == NULL:
            return -2

        io_uring_prep_recv(sqe, sockfd, c_buffer, length, flags)
        self.set_data(sqe, user_data)
        return 1

    cpdef int schedule_send(self, object user_data, int sockfd, bytes buf, int flags=0):
        if self.__check(io_uring_op.IORING_OP_SEND) == 0:
            return -1
        cdef char* c_buffer = buf
        cdef int length = len(buf)
        cdef io_uring_sqe* sqe = self.__get_sqe()
        if sqe == NULL:
            return -2

        io_uring_prep_send(sqe, sockfd, c_buffer, length, flags)
        self.set_data(sqe, user_data)
        return 1

    cpdef int schedule_socket(self, object user_data, int domain, int type, int protocol, unsigned int flags=0):
        if self.__check(io_uring_op.IORING_OP_SOCKET) == 0:
            return -1
        cdef io_uring_sqe* sqe = self.__get_sqe()
        if sqe == NULL:
            return -2

        io_uring_prep_socket(sqe, domain, type, protocol, flags)
        self.set_data(sqe, user_data)
        return 1

    cpdef int schedule_splice(self, object user_data, int src, int dst, unsigned int length, long off_in=-1, long off_out=-1, unsigned int flags=0):
        if self.__check(io_uring_op.IORING_OP_SPLICE) == 0:
            return -1
        cdef io_uring_sqe* sqe = self.__get_sqe()
        if sqe == NULL:
            return -2

        io_uring_prep_splice(sqe, src, off_in, dst, off_out, length, flags)
        self.set_data(sqe, user_data)
        return 1

    cpdef int schedule_cancel(self, object user_data, object other_user_data, int flags=0):
        if self.__check(io_uring_op.IORING_OP_ASYNC_CANCEL) == 0:
            return -1
        cdef io_uring_sqe* sqe = self.__get_sqe()
        if sqe == NULL:
            return -2

        io_uring_prep_cancel(sqe, <void*>other_user_data, flags)
        self.set_data(sqe, user_data)
        return 1

    cpdef submit(self):
        if self.open:
            return io_uring_submit(&self.ring)


    