import os
import types
import asyncio
import warnings

from .core import IoUring, SockAddr, StatxResult

cdef int AT_FDCWD = -100
cdef int AT_EMPTY_PATH  = 0x1000

class IoUringPlugin(object):
    def __new__(self, loop):
        plugin = self.init_plugin(loop)
        return plugin

    def init_plugin(loop):
        return ReaderEventLoopPlugin(loop)

def on_ring_events(plugin):
    #print("ON RING EVENTS", plugin)
    result = plugin.ring.get_completion()
    while result is not None:
        data = result.data
        if result.result < 0:
            data.set_exception(OSError(-result.result))
        else:
            data.set_result(result.result)
        result = plugin.ring.get_completion()

def loop_close(self):
    if hasattr(self, '_io_uring_plugin'):
        plugin = self._io_uring_plugin
        self.remove_reader(plugin.ring_eventfd)
        delattr(self, '_io_uring_plugin')
        plugin.close()
    self._io_uring_plugin_close()


class ReaderEventLoopPlugin(EventLoopPlugin):
    
    def __init__(self, loop):
        super().__init__()
        self.loop = loop
        loop.add_reader(self.ring_eventfd, on_ring_events, self)
        setattr(loop, '_io_uring_plugin', self)

        # Maybe there is a better way to hook the event loops close method?

        loop._io_uring_plugin_close = loop.close
        loop.close = types.MethodType(loop_close, loop)

        

class OperationNotSupportedError(Exception):
    pass

class RingFullError(Exception):
    pass


cdef class EventLoopPlugin:

    def __init__(self):
        self.ring = IoUring(entries=256)
        self.ring_eventfd = os.eventfd(0)
        self.ring.register_eventfd(self.ring_eventfd)

    async def aclose(self):
        self.close()

    def close(self):
        self.ring.close()
        os.close(self.ring_eventfd)
        self.ring_eventfd = -1

    def closed(self):
        return self.ring_eventfd == -1
        
    def __del__(self):
        if not self.closed():
            warnings.warn("Unclosed EventLoopPlugin detected", ResourceWarning)
            self.close()

    def _handle_res(self, result, future):
        if result == -1:
            future.set_exception(OperationNotSupportedError())
        elif result == -2:
            future.set_exception(RingFullError())
        else:
            self.ring.submit()

    async def sock_sendfile(self, sock_fd, file_fd, offset=0, count=None, fallback=True, pipe=None):
        buffer_size = 65536 # default linux pipe size
        sent = 0
        
        if not pipe:
            pipe = os.pipe()
        pipe_out, pipe_in = pipe
        if not count:
            count =(await self.file_fstat(file_fd)).st_size - offset
        
        while count > sent:
            bytes_to_transfer = min(count, buffer_size)
            transfered = await self._splice(file_fd, pipe_in, bytes_to_transfer, offset + sent, -1)
            if transfered == 0:
                break
            sent += transfered
            await self._splice(pipe_out, sock_fd, transfered)

        os.close(pipe_out)
        os.close(pipe_in)
        return sent

        
    async def _splice(self, fd_in, fd_out, length, off_in = -1, off_out = -1):
        future = self.loop.create_future()
        res = self.ring.schedule_splice(future, fd_in, fd_out, length, off_in, off_out)
        self._handle_res(res, future)
        return await future
        
    async def file_open(self, path, flags=0, mode=0o644):
        future = self.loop.create_future()
        res = self.ring.schedule_openat(future, AT_FDCWD, os.fsencode(path), flags, mode)
        self._handle_res(res, future)
        return await future
    
    async def file_close(self, fd):
        future = self.loop.create_future()
        res = self.ring.schedule_close(future, fd)
        self._handle_res(res, future)
        await future
        return True
    
    async def file_read(self, fd, n, offset):
        buffer = bytes(n)
        await self.file_read_into(fd, buffer, offset)
        return buffer

    async def file_read_into(self, fd, buffer, offset):
        future = self.loop.create_future()
        res = self.ring.schedule_read(future, fd, buffer, offset)
        self._handle_res(res, future)
        return await future

    async def file_write(self, fd, data, offset):
        future = self.loop.create_future()
        res = self.ring.schedule_write(future, fd, data, offset)
        self._handle_res(res, future)
        return await future

    async def file_stat(self, path):
        statx = StatxResult()
        future = self.loop.create_future()
        res = self.ring.schedule_statx(future, AT_FDCWD, os.fsencode(path), statx, 0)
        self._handle_res(res, future)
        await future
        return statx

    async def file_fstat(self, fd):
        statx = StatxResult()
        future = self.loop.create_future()
        res = self.ring.schedule_statx(future, fd, b'', statx, flags = AT_EMPTY_PATH)
        self._handle_res(res, future)
        await future
        return statx
